"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Cobbler = exports.BaseCobbler = void 0;
const gdut_json_1 = require("gdut-json");
const gdut_io_1 = require("gdut-io");
const cobbler_handler_1 = require("./cobbler-handler");
function BaseCobbler(path = './cobbler') {
    const tables = gdut_json_1.ReadJson(`${path}/tables.json`);
    return new Cobbler(`${path}/data`, tables);
}
exports.BaseCobbler = BaseCobbler;
class Cobbler {
    constructor(dataPath, tables, handler = new cobbler_handler_1.CobblerHandler(), cache = true) {
        this.dataPath = dataPath;
        this.tables = tables;
        this.handler = handler;
        this.cache = cache;
        this.tableCache = {};
    }
    getTables() {
        return Object.entries(this.tables).map(([key, value]) => {
            return Object.assign({ tableName: key }, value);
        });
    }
    saveAll(name, records) {
        return records.map((record) => this.save(name, record));
    }
    save(name, record) {
        const records = this.readDataFile(name);
        let saved = null;
        if (record.id && record.id > 0) {
            const found = this.locate(record.id, records);
            if (!found) {
                saved = this.callInsert(name, records, record, saved);
            }
            else {
                if (found.logicRemove) {
                    throw new Error(`RECORD_REMOVED; Record "${found.id}" already Removed! 🤯`);
                }
                saved = this.callUpdate(name, records, record, found, saved);
            }
        }
        else {
            saved = this.callInsert(name, records, record, saved);
        }
        this.writeDataFile(name, records);
        return saved;
    }
    callUpdate(name, records, record, found, saved) {
        if (this.handler.beforeUpdate(name, records, record)) {
            saved = this.update(name, records, found, record);
            this.handler.afterUpdate(name, records, record, saved);
        }
        return saved;
    }
    callInsert(name, records, record, saved) {
        if (this.handler.beforeInsert(name, records, record)) {
            saved = this.insert(name, records, record);
            this.handler.afterInsert(name, records, record, saved);
        }
        return saved;
    }
    remove(name, id) {
        const records = this.readDataFile(name);
        const found = this.locate(id, records);
        if (found) {
            if (this.handler.beforeRemove(name, records, found)) {
                found.logicRemove = true;
                this.writeDataFile(name, records);
                this.handler.afterRemove(name, records, found);
            }
            return found;
        }
        else {
            throw new Error(`RECORD_NOT_FOUND; Record not found ${id}! 😵`);
        }
    }
    findById(name, id) {
        const records = this.readDataFile(name);
        const found = this.locate(id, records);
        if (found) {
            return this.transformResult(name, found);
        }
        else {
            return null;
        }
    }
    findAll(name, parms = {}) {
        const records = this.readDataFile(name);
        return records
            .filter((el) => !el.logicRemove)
            .filter((el) => this.columnFilter(name, parms, el))
            .map((el) => this.transformResult(name, el))
            .filter((el) => this.handler.filter(name, el, parms));
    }
    columnFilter(name, parms, record) {
        let valid = true;
        const tableConfig = this.tables[name];
        const config = Object.assign(Object.assign({}, tableConfig), { id: { type: 'number' } });
        Object.entries(config).forEach(([key, columnConfing]) => {
            if (valid) {
                valid = this.invokeColumnFilter(key, columnConfing, parms, record);
            }
        });
        return valid;
    }
    invokeColumnFilter(name, column, parms, record) {
        let valid = true;
        if (typeof parms[`${name}_equal`] === column.type) {
            valid = parms[`${name}_equal`] === record[name];
        }
        if (typeof parms[`${name}_not_equal`] === column.type) {
            valid = parms[`${name}_not_equal`] !== record[name];
        }
        if (typeof parms[name] === column.type) {
            valid = parms[name] === record[name];
        }
        if (column.type === 'string') {
            if (typeof parms[`${name}_like`] === column.type) {
                valid = record[name].includes(parms[`${name}_like`]);
            }
            if (typeof parms[`${name}_not_like`] === column.type) {
                valid = !record[name].includes(parms[`${name}_not_like`]);
            }
        }
        if (column.type === 'number') {
            if (typeof parms[`${name}_greater`] === 'string') {
                const val = parseInt(parms[`${name}_greater`]);
                valid = record[name] > val;
            }
            if (typeof parms[`${name}_less`] === 'string') {
                const val = parseInt(parms[`${name}_less`]);
                valid = record[name] < val;
            }
        }
        return valid;
    }
    findPage(name, page, size, sortName, sortDir, parms) {
        const records = this.findAll(name, parms);
        const count = records.length;
        const items = records
            .sort(CompareValues(sortName, sortDir))
            .slice(page * size, page * size + size);
        return { count, items };
    }
    transformResult(name, source) {
        const record = Object.assign({}, source);
        this.transformRecord(name, record);
        this.collectOwnedReferences(name, record);
        return record;
    }
    transformRecord(name, record, parent) {
        const tableConfig = this.tables[name];
        Object.keys(tableConfig).forEach((columnName) => {
            const columnConfig = tableConfig[columnName];
            const value = record[columnName];
            if (value !== null && columnConfig.reference && columnConfig.reference !== parent) {
                const refs = this.readDataFile(columnConfig.reference);
                record[columnName] = refs.find((el) => el.id === value);
            }
        });
        return record;
    }
    collectOwnedReferences(name, record) {
        Object.keys(this.tables)
            .filter((tableName) => tableName !== name)
            .forEach((tableName) => {
            const tableConfig = this.tables[tableName];
            Object.entries(tableConfig).filter(([colName, colConfig]) => {
                if (colConfig.reference === name && colConfig.owner) {
                    const refs = this.readDataFile(tableName);
                    record[tableName] = refs
                        .filter((ref) => !ref.logicRemove && ref[colName] === record.id)
                        .map((el) => {
                        this.transformRecord(tableName, el, name);
                        return el;
                    });
                }
            });
        });
    }
    prepareOwnedReferences(name, record) {
        const out = {};
        Object.keys(this.tables)
            .filter((tableName) => tableName !== name)
            .forEach((tableName) => {
            const tableConfig = this.tables[tableName];
            Object.entries(tableConfig).filter(([colName, colConfig]) => {
                if (colConfig.reference === name && colConfig.owner) {
                    if (record[tableName]) {
                        out[tableName] = record[tableName];
                        delete record[tableName];
                    }
                }
            });
        });
        return out;
    }
    getReferenceColumnName(childName, parentName) {
        const tableConfig = this.tables[childName];
        let reference = null;
        Object.keys(tableConfig).forEach((columnName) => {
            const columnConfig = tableConfig[columnName];
            if (columnConfig.reference === parentName) {
                reference = columnName;
            }
        });
        if (reference) {
            return reference;
        }
        else {
            throw new Error(`Reference ${parentName} not found in ${childName}`);
        }
    }
    saveOwnedReferences(name, parent, prepared) {
        Object.keys(prepared).forEach((tableName) => {
            const records = prepared[tableName];
            const parentColumn = this.getReferenceColumnName(tableName, name);
            if (Array.isArray(records) && records.length > 0) {
                records.forEach((record) => {
                    record[parentColumn] = parent.id;
                    this.save(tableName, record);
                });
            }
        });
    }
    update(name, records, found, record) {
        const referenceData = this.prepareOwnedReferences(name, record);
        // UPDATE Found Record! - Mutable mode!
        Object.keys(record).forEach((key) => {
            if (!['id', 'logicRemove'].includes(key)) {
                found[key] = record[key];
            }
        });
        this.validate(name, records, found, 'UPDATE');
        this.saveOwnedReferences(name, found, referenceData);
        return found;
    }
    insert(name, records, record) {
        const referenceData = this.prepareOwnedReferences(name, record);
        if (typeof record.id === 'undefined' || record.id < 1) {
            record.id = records.reduce((p, c) => (p > c.id ? p : c.id), 0) + 1;
        }
        record.logicRemove = false;
        this.validate(name, records, record, 'INSERT');
        records.push(record);
        this.saveOwnedReferences(name, record, referenceData);
        return record;
    }
    validate(name, records, record, action) {
        if (this.tables[name]) {
            const columns = this.tables[name];
            Object.keys(columns).forEach((columnName) => {
                const columnConfig = columns[columnName];
                const value = record[columnName];
                const valueType = typeof record[columnName];
                this.checkColumnType(valueType, columnConfig, record, columnName, value);
                this.checkReference(columnConfig, columnName, value);
                this.checkUnique(record, records, columnName, columnConfig, value);
            });
        }
    }
    checkUnique(record, records, columnName, columnConfig, value) {
        if (columnConfig.unique) {
            const found = records.find((el) => el.id !== record.id && el[columnName] === value);
            if (found) {
                throw new Error(`UNIQUE_CONFLICT; The "${columnName}" / [${value}] is not unique - id:${found.id}! 😬`);
            }
        }
    }
    checkColumnType(valueType, columnConfig, record, columnName, value) {
        if (valueType === 'undefined' || value === null) {
            if (typeof columnConfig.default !== 'undefined') {
                record[columnName] = columnConfig.default;
                valueType = typeof record[columnName];
            }
            else if (columnConfig.needed) {
                throw new Error(`COLUMN_NEEDED; Column needed: "${columnName}" ! 🤨`);
            }
            else {
                record[columnName] = null;
                value = null;
            }
        }
        if (valueType !== columnConfig.type && value !== null) {
            throw new Error(`COLUMN_TYPE_MISMATCH; Column "${columnName}" type "${columnConfig.type}" mismatch "${valueType}" ! 😲`);
        }
    }
    checkReference(columnConfig, columnName, value) {
        if (columnConfig.reference && value !== null) {
            if (this.tables[columnConfig.reference]) {
                this.locateReference(columnConfig.reference, columnName, value);
            }
            else {
                throw new Error(`Column "${columnName}" - reference table not found: ${columnConfig.reference}!`);
            }
        }
    }
    locateReference(reference, columnName, value) {
        const refRecs = this.readDataFile(reference);
        const refFound = refRecs.find((el) => el.id === value);
        if (!refFound) {
            throw new Error(`Column "${columnName}" - reference record not found ${reference} / ${value}!`);
        }
        else if (refFound.logicRemove) {
            throw new Error(`Column "${columnName}" - reference record deleted ${reference} / ${value}!`);
        }
    }
    locate(id, data) {
        const found = data.find((rec) => rec.id === id);
        if (found) {
            return found;
        }
        else {
            return null;
        }
    }
    readDataFile(name) {
        if (this.cache && this.tableCache[name]) {
            return this.tableCache[name];
        }
        else {
            const filePath = `${this.dataPath}/${name}.json`;
            if (gdut_io_1.FileExists(filePath)) {
                const data = gdut_json_1.ReadJson(filePath);
                if (this.cache) {
                    this.tableCache[name] = data;
                }
                return data;
            }
            else {
                return [];
            }
        }
    }
    writeDataFile(name, data) {
        const filePath = `${this.dataPath}/${name}.json`;
        if (this.cache) {
            this.tableCache[name] = data;
        }
        gdut_json_1.WriteJson(filePath, data, true);
    }
}
exports.Cobbler = Cobbler;
function CompareValues(key, order = 'ASC') {
    return function innerSort(a, b) {
        if (!a.hasOwnProperty(key) || !b.hasOwnProperty(key)) {
            // property doesn't exist on either object
            return 0;
        }
        const varA = typeof a[key] === 'string' ? a[key].toUpperCase() : a[key];
        const varB = typeof b[key] === 'string' ? b[key].toUpperCase() : b[key];
        let comparison = 0;
        if (varA > varB) {
            comparison = 1;
        }
        else if (varA < varB) {
            comparison = -1;
        }
        return order === 'DESC' ? comparison * -1 : comparison;
    };
}
